;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;                                               ;;;
;;;  Subtraction & Typing Model with support      ;;;
;;;                                               ;;;
;;;  Jelmer Borst (jpborst@ai.rug.nl)             ;;;
;;;  Department of Artificial Intelligence        ;;;
;;;  University of Groningen                      ;;;
;;;  The Netherlands                              ;;;
;;;                                               ;;;
;;;  20120229                                     ;;;
;;;                                               ;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  
;;;  
;;; 

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defparameter *realtime* nil)
(defparameter *modelexpdone* nil)

(defparameter *model-pathname* ;pathname to outputfile, default "pp-test"
  (merge-pathnames 
   (make-pathname :name "model" :directory '(:relative "modeloutput") :type "dat")  
   (current-pathname)))



(setf *model* t)


;;; Code to run the experiment with the model
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;; run some trials
(defun m ()

  ;(reset-experiment)

  (reset)

  ;(threads-reset (get-module threads))

  (setf *output-pathname* *model-pathname*)
  (setf *eye* nil)
  (setf *pp* "model1")
  (setf *model* t)
  (setf *realtime* nil)

  (setf *exp-part* "real")

  (setf *trial* 0)

  (setf *signal* t)

  (setf *current-levels*
        '(
         
          (hard hard)
      
          ))

  (mp:process-run-function "experiment" nil 'e "real"))


;; run some full models
(defun mX (n &optional (pp 1))

  (setf *output-pathname* *model-pathname*)
  (setf *model* t)
  (setf *eye* nil)
  (setf *realtime* nil)
  (setf *modelexpdone* nil) 

  (dotimes (i n)

    (reset-experiment)
    (setf *eye* nil)
    (setf *current-levels* '())
    (setf *trial* 0)

    (dotimes (j 2)

      (if (equalp j 0)
          (progn
            (setf *signal* (if (oddp n) t nil))
            (setf *exp-part* "real1"))
          (progn
            (setf *signal* (if (oddp n) nil t))
            (setf *exp-part* "real2")))

      
      (setf *current-levels* (get-block))
      (reset)
      
      (setf *pp* (format nil "model~a" (+ i pp)))
      
      (e (format nil "~a" *exp-part*))
      
      (mp:process-wait "wait-for-exp" (lambda () *modelexpdone*))
      
      (setf *modelexpdone* nil)
      
      (gc-if-needed)
      (sleep 15))))




(defun reset-experiment ()
  (setf *typing-stimuli-hard*
        (randomize '("zorgvuldig"
                     "gezelschap"
                     "terugkomen"
                     "uitstekend"
                     "regelmatig"
                     "ruimtelijk"
                     "bewustzijn"
                     "bevestigen"
                     "instelling"
                     "literatuur"
                     "overtuigen"
                     "gewoonlijk"
                     "aanleiding"
                     "historisch"
                     "beschermen"
                     "gevaarlijk"
                     "toepassing"
                     "voorwaarde"
                     "fluisteren"
                     "ziekenhuis"
                     "verklaring"
                     "onmogelijk"
                     "beslissing"
                     "activiteit"
                     "schreeuwen"
                     "informatie"
                     "behandelen"
                     "verhouding"
                     "wetenschap"
                     "vervolgens"
                     "economisch"
                     "betrekking"
                     "glimlachen"
                     "beschouwen"
                     "nauwelijks"
                     "plotseling"
                     "antwoorden"
                     "veranderen"
                     "verdwijnen"
                     "belangrijk"
                     "natuurlijk"
                     )))

  ;;stimuli for easy typing, scrambled hard stimuli without repeating same character
  (setf *typing-stimuli-easy*
        (randomize '("gluozidgvr"
                     "tiusbejzwn" 
                     "winjervedn" 
                     "suteilrfen" 
                     "emgerkuotn" 
                     "iocnocshem" 
                     "sbesinsgil" 
                     "iesaonpsgt" 
                     "ebnlehnead" 
                     "klsjweniau" 
                     "uarluijtnk" 
                     "iktemlruji" 
                     "ilgrbnjaek" 
                     "nrougdivhe" 
                     "inksiezhue" 
                     "zceahglsep" 
                     "oegnvlvrse" 
                     "cserenhmeb" 
                     "iaticivtet" 
                     "aigegtrmel" 
                     "adilaenign" 
                     "wenshatcpe" 
                     "esbeinvetg" 
                     "nioitfemar" 
                     "avrarodweo" 
                     "mleialghcn" 
                     "aedtornown" 
                     "tsoirihshc" 
                     "tuiretlura" 
                     "ptlsnloieg" 
                     "engiolwjok" 
                     "rktekinebg" 
                     "esohcbuwne" 
                     "dtenisektu" 
                     "uercweshen" 
                     "iruveontge" 
                     "nljmoigkeo" 
                     "instlielng" 
                     "elkrivarng" 
                     "naerdvenre"
                     "valragijek" )))
  
  ;;easy subtraction stimuli
  
  (setf *subtraction-stimuli-easy*
        (get-easy-stims 41))

  ;;hard subtraction stimuli

  (setf *subtraction-stimuli-hard*
        (get-hard-stims 41))
  (format t "experiment resetted~%"))




;;; Modeling functions
;;;;;;;;;;;;;;;;;;;;;;;

(defun find-location-x (s)
  "Hack to find a button with string s"   ;now really only BUTTONS!!
  (let ((answer nil)
        (l (visicon-chunks (get-module :vision))))
    (while (and (null answer) l)
        (if (and
             (equal (chunk-slot-value-fct (first l) 'kind) 'BUTTON) 
             (equal (string-upcase s)(string-upcase (chunk-real-visual-value (first l)))))
            (setf answer (first l))
          (setf l (Rest l))))
      (if answer (chunk-slot-value-fct answer 'screen-x)
        (progn (format t "~%*** Warning: find-location-x called with ~S which is not on the screen ***~%" s)
          nil))))
  
(defun find-location-y (s)
  "Hack to find a button with string s"   ;now really only BUTTONS!!
  (let ((answer nil)
        (l (visicon-chunks (get-module :vision))))
    (while (and (null answer) l)
        (if (and
             (equal (chunk-slot-value-fct (first l) 'kind) 'BUTTON) 
             (equal (string-upcase s)(string-upcase (chunk-real-visual-value (first l)))))
            (setf answer (first l))
          (setf l (Rest l))))
      (if answer (chunk-slot-value-fct answer 'screen-y)
        (progn (format t "~%*** Warning: find-location-y called with ~S which is not on the screen ***~%" s)
          nil))))

(defun text-not-attended ()
  (awhen (get-module :vision)  ;; Test that there is a vision module
         (update-new it)
         (check-finsts it)) 
  (let ((l (visicon-chunks (get-module :vision))))
    (dolist (element l)
      (when (and 
             (equalp (chunk-slot-value-fct element 'kind) 'TEXT)
             (not (equalp (feat-attended element (get-module :vision)) t)))
        (return t))))) 

(defun look-right ()
  (if (> (svref (eye-loc (get-module :vision)) 0) 500) t nil))

;;; Disable buffer stuffing
(defmethod stuff-visloc-buffer ((vis-mod vision-module)) nil)

;currently requested chunk
(defparameter *cur-request* nil)

;;; calc-simularities for partial matching
(defun calc-similarities (a b)
  (let ((result nil))
    (when (and (numberp a) (numberp b)) ;order facts & problem states
      (let ((a (if (not (numberp a)) (parse-integer a :junk-allowed t) a))
            (b (if (not (numberp b)) (parse-integer b :junk-allowed t) b)))
        (when (and a b)
          (when (equalp (chunk-spec-chunk-type *cur-request*) 'typing-word)
            (setf result (* .01 (- (sqr (- a b)))))
            )
          (when (equalp (chunk-spec-chunk-type *cur-request*) 'typing-pos)
            (setf result (* .8 (- (sqr (- a b)))))
            )

          (when (equalp (chunk-spec-chunk-type *cur-request*) 'carry-subtraction)
            (setf result (* .03 (- (sqr (- a b)))))
            )
          (when (equalp (chunk-spec-chunk-type *cur-request*) 'order)
            (setf result (* (max 1 (- 1 (/ a 10))) (- (sqr (- a b))))))))) 
    
    (when (not (and (numberp a) (numberp b))) 
      (let ((a (parse-integer a :junk-allowed t))
            (b (parse-integer b :junk-allowed t)))
        (when (and a b)
          (if (< a 11) (setf result (* (max .9 (- 1 (/ a 20))) (- (sqr (- a b))))))
          (if (> a 10) (setf result (* (max .5 (- 1 (/ a 20))) (- (sqr (- a b))))))
          (if (or (and (> a 10) (< b 11)) (and (< a 10) (> b 11))) (setf result -100)))))
    result))

(defun my-partial-matching (chunk request)
  (setf *cur-request* request)
  (let ((total-sim 0.0)
        (dm (get-module declarative)))
    (dolist (k (chunk-spec-slot-spec request))
      (when (and (or (eq (car k) '=)
                     (eq (car k) '-))
                 (not (chunk-spec-variable-p (third k))))
        (when (dm-act-level (dm-act dm) 'medium)
          (model-output "  comparing slot ~S"
                        (second k))
          (model-output "  Requested: ~s ~s  Chunk's slot value: ~s"
                        (first k)
                        (third k) 
                        (fast-chunk-slot-value-fct 
                         chunk
                         (second k))))
        
        (let* ((sim (chunks-similarity dm 
                                       (third k) 
                                       (fast-chunk-slot-value-fct chunk (second k))
                                       t))
               
               
               (sim-dif (case (car k) 
                          (= (* (dm-mp dm) sim))
                          
                          (- (cond ((= sim (dm-ms dm))
                                    (* (dm-mp dm) (dm-md dm)))
                                                  ; ACT-R 5 doesn't do this, but
                                                  ; it seems like maybe it should
                                                  ;((= sim (dm-md dm))
                                                  ; (* (dm-mp dm) (dm-ms dm)))
                                   (t
                                    (when (dm-act-level (dm-act dm) 'medium)
                                      (model-output 
                                       "  negation test with similarity not ms has no effect"))
                                    0))))))
          
          (when (dm-act-level (dm-act dm) 'medium)
            (model-output "  effective similarity value is ~f" sim-dif))
          
          (incf total-sim sim-dif)
          
          )))
    (when (dm-act-level (dm-act dm) 'medium)
      (model-output "Total similarity score ~f" total-sim))
    
    


    (when (equalp (chunk-spec-chunk-type request) 'sub-fact) ;positive and negative outcomes are very dissimilar! even if the slots are more or less the same.
      (let 
          ((diff-request 
            (- 
             (parse-integer (third (first (chunk-spec-slot-spec request 'terma)))) 
             (parse-integer (third (first (chunk-spec-slot-spec request 'termb))))))
           (diff-chunk
            (-
             (parse-integer (fast-chunk-slot-value-fct chunk 'terma))
             (parse-integer (fast-chunk-slot-value-fct chunk 'termb)))))

       ; (print (chunk-spec-chunk-type request))
       ; (print (chunk-spec-slot-spec request))
       ; (print diff-request)
       ; (format t "~%terma: ~a, termb: ~a, diff: ~a" (fast-chunk-slot-value-fct chunk 'terma) (fast-chunk-slot-value-fct chunk 'termb) diff-chunk)
       ; (format t "~%")
        (when (or 
               (and (> diff-request -1) (< diff-chunk 0)) 
               (and (< diff-request 0) (> diff-chunk -1)))
          (decf total-sim 100)
         ; (format t "not similar! ~a" total-sim)
          )))
    
    total-sim))


;; my reset - just clearing out buffers, emptying dm, but NOT emptying trace/resetting time
(defun model-reset ()
  (verify-current-mp  
   "reset called with no current meta-process."
    (let ((mp (current-mp)))
     ;(reset-mp mp)
      (setf (meta-p-events mp) nil)
      (setf (meta-p-delayed mp) nil)
      
      (maphash #'(lambda (name model)
                   (declare (ignore name))
                   (reset-my-model mp model))
               (meta-p-models mp)))))


(defun reset-my-model (mp model)
 
  (let ((previous-model (meta-p-current-model mp)))
    (setf (meta-p-current-model mp) model)
    
    (clrhash (act-r-model-chunk-types-table model))
    (clrhash (act-r-model-chunks-table model))
    (clrhash (act-r-model-chunk-ref-table model))
    
    (setf (act-r-model-chunk-update model) t)
    (setf (act-r-model-dynamic-update model) t)
    (setf (act-r-model-delete-chunks model) nil)
    
    (setf (act-r-model-largest-chunk-type model) 0)
    
    (maphash #'(lambda (buffer-name buffer)
                 (declare (ignore buffer-name))
                 (setf (act-r-buffer-chunk buffer) nil))
             (act-r-model-buffers model))
    
    (create-model-default-chunk-types-and-chunks)
    
    (maphash #'(lambda (module-name instance)
                 ;(print module-name))
                 (declare (ignore instance))
                  (if (not (equalp module-name 'buffer-trace)) (reset-module module-name)))
             (global-modules-table))
    
    (maphash #'(lambda (parameter-name parameter)
                 (sgp-fct (list parameter-name (act-r-parameter-default parameter))))
             *act-r-parameters-table*)
    
    
    (maphash #'(lambda (module-name val)
                 (declare (ignore val))
                  (if (not (equalp module-name 'buffer-trace)) (secondary-reset-module module-name)))
             (global-modules-table))    
    
    (dolist (form (act-r-model-code model))
      (eval form))
    
    (maphash #'(lambda (module-name val)
                 (declare (ignore val))
                  (if (not (equalp module-name 'buffer-trace))(tertiary-reset-module module-name)))
             (global-modules-table))
    
    (setf (meta-p-current-model mp) previous-model)))



;;; Model
;;;;;;;;;;;;;;;;;

(clear-all)

(define-model subtractionTyping



(sgp :esc t
     :er t
     :egs 0
     :bll .5
     :rt -3.0
     :lf .3
     :ans .1 
     :ol t 
     :crt nil
     :cst nil
     :mp 1 ;1 ;partial matching
     :ms 0 ;maximum similarity
     :md -1  ;maximum difference
     :sim-hook calc-similarities
     :partial-matching-hook my-partial-matching

     :needs-mouse nil
     :show-focus nil    
     :optimize-visual t  
     :randomize-time nil
     :visual-movement-tolerance 0 
     :process-cursor nil
     :v nil
     :trace-detail low 
     :ga 1
     :act nil
     :declarative-finst-span 3
     :mas nil
     :BUFFER-TRACE nil
     :SAVE-BUFFER-TRACE nil
     :save-dm-history nil
     :ncnar nil
     
     :SACCADE-BASE-TIME           0.003  
     :EYE-SACCADE-RATE            0.001  
     :VISUAL-ENCODING-EXPONENT    .15       
     :VISUAL-ATTENTION-LATENCY    0.065    
     :VISUAL-ENCODING-FACTOR      0.002   
     
     :MOUSE-FITTS-COEFF .085 
     )


(set-hand-location left 21 5)
(set-hand-location right 28 2)

;;; General
;;;;;;;;;;;;;;;

;;DM

(chunk-type typing state last id step vis-sp)
(chunk-type subtraction type state last id step)




(chunk-type (button (:include visual-object)) tag)
(chunk-type (circle (:include visual-object)) tag)
(chunk-type (option (:include visual-object)) tag)
;(chunk-type (text (:include visual-object)) x y)

(add-dm
 (start isa chunk)
 (enter isa chunk)
 (enter-easy isa chunk)
 (enter-hard isa chunk)
 (next-letter isa chunk)
 (easy isa chunk)
 (next-number isa chunk)
 (next-number-easy isa chunk)
 (next-number-hard isa chunk)
 (restore isa chunk)
 (circle isa chunk)
 (button isa chunk)
 (option isa chunk)
 (grey isa chunk)
 (listening isa chunk)
 (reply isa chunk)
 (add-tens isa chunk)
 (next-word isa chunk)
 (encode-carry isa chunk)
 (almost-next-letter isa chunk)
 (busy-subtracting isa chunk)


; (typing-goal isa typing state start)
; (subtraction-goal isa subtraction type easy state start)

 ;(keyboard-right isa visual-location kind button screen-x 790 screen-y 588 width 350 height 150)
 ;(keyboard-left isa visual-location kind button screen-x 195 screen-y 588 width 150 height 150)
 (task-right isa visual-location kind button screen-x 690 screen-y 190 width 20 height 20)
 (task-left isa visual-location kind button screen-x 25 screen-y 135 width 20 height 20)
 ;(task-carry-left isa visual-location kind button screen-x 235 screen-y 100 width 20 height 20)
)


;;; Typing
;;;;;;;;;;;;;;;

;;DM

(chunk-type typing-word word step id)
(chunk-type typing-pos character pos step id)
(chunk-type order identity next)
(chunk-type carry-subtraction carry termB step id)

(add-dm
 (a isa order identity 0 next 1)
 (b isa order identity 1 next 2)
 (c isa order identity 2 next 3)
 (d isa order identity 3 next 4)
 (e isa order identity 4 next 5)
 (f isa order identity 5 next 6)
 (g isa order identity 6 next 7)
 (h isa order identity 7 next 8)
 (i isa order identity 8 next 9)
 (j isa order identity 9 next 10)
 (k isa order identity 10 next 11)
 )

(sdp a :creation-time -1e12 :references 3500000) 
(sdp b :creation-time -1e12 :references 3500000) 
(sdp c :creation-time -1e12 :references 3500000)
(sdp d :creation-time -1e12 :references 3500000) 
(sdp e :creation-time -1e12 :references 3500000)
(sdp f :creation-time -1e12 :references 3500000) 
(sdp g :creation-time -1e12 :references 3500000) 
(sdp h :creation-time -1e12 :references 3500000)
(sdp i :creation-time -1e12 :references 3500000) 
(sdp j :creation-time -1e12 :references 3500000) 
(sdp k :creation-time -1e12 :references 3500000)

;;Production Rules



;;circle processing
(p if-circles-look-at-them
  ?goal>
    buffer empty
  ?visual-location>
    buffer empty

 !eval! (not (equalp *sub-circle-color* :white))
==>
  +visual-location>
    isa visual-location
    kind circle
  < screen-x 512
)

(p attend-left-circle-green

  =visual-location>
    isa visual-location
    kind circle
    color green
  < screen-x 512
==>
  +visual-location>
    isa visual-location
    kind circle
  > screen-x 512
  +goal>
    isa subtraction
    type easy
    state start
  +goal>
    isa typing
    state start)

(p attend-left-circle-red

  =visual-location>
    isa visual-location
    kind circle
    color red
  < screen-x 512
==>
  +visual-location>
    isa visual-location
    kind circle
  > screen-x 512
  +goal>
    isa subtraction
    type hard
    state start
  +goal>
    isa typing
    state start)


(p attend-right-circle
  =visual-location>
    isa visual-location
    kind circle
  > screen-x 512
  ?visual>
    state free
==>
  +visual>
    isa clear
)




;;TYPING

(p typing-start
   !eval! (typing-stimulus *exp*)
   !eval! *typing-on*
   =goal>
     isa typing
     state start
   ?visual-location>
     buffer empty
   - state error
   ?visual>
     buffer empty
     state free
   ?imaginal>
     state free
   ?retrieval>
     state free
  ; ?manual>
  ;   state free
==>
  =goal>
    state enter
  +visual-location>
    isa visual-location
    kind text
    > screen-x 500
    < screen-x 1000
    screen-y 190
)

(p attend-stimulus
   !eval! *typing-on*
   =goal>
     isa typing
     state enter
   =visual-location>
     isa visual-location
     kind text
   > screen-x 510
   ?visual>
     state free
     buffer empty
==>
  +visual>
    isa move-attention
    screen-pos =visual-location
)


;;;easy typing


(p direct-click-initiate
   !eval! *typing-on*
   !eval! (look-right)
   =goal>
     isa typing
     state enter
   - last =txt
   =visual>
     isa text
     value =txt
   !eval! (equalp (length =txt) 1)
==>
  =goal>
    state enter-easy
   !bind! =x (find-location-x =txt)
   !bind! =y (find-location-y =txt)
   +visual-location>
     isa visual-location
     kind button
     screen-x =x
     screen-y =y
)

(p direct-click-find
   !eval! *typing-on*
   =goal>
     isa typing
     state enter-easy
   =visual-location>
     isa visual-location
     kind button
   ?visual>
     state free
     buffer empty
   ?manual>
     state free
==>
  +visual>
    isa move-attention
    screen-pos =visual-location
  +manual>
    isa move-cursor
    loc =visual-location
)

(p direct-click-click
   !eval! *typing-on*
  =goal>
    isa typing
    state enter-easy
  =visual>
    isa button
    value =txt
  ?visual>
    state free
  ?manual>
    state free
 ==>
   =goal>
     state busy-subtracting
     last =txt
   +manual>
     isa click-mouse
   +visual>
     isa clear
)


(p switch-to-typing-start
   !eval! *subtraction-on*
  =goal>
     isa typing
     state busy-subtracting
  ?imaginal>
     state free
==>
  =goal>
    state start
)

;;;hard typing

(p hard-click-initiate
   !eval! *typing-on*
   =goal>
     isa typing
     state enter
   =visual>
     isa text
     value =txt
   !eval! (> (length =txt) 1)
==>
  !bind! =id *trial* ; (random 10000)
  !bind! =step *typ_step*
  =goal>
    state get-pos
    id =id
    step =step
    vis-sp =txt
  +imaginal>
    isa typing-word
    word =txt
    id =id
    step =step
)

(p start-typing-position-hard
 !eval! *typing-on*
   =goal>
     isa typing
     state get-pos
     step =step
     id =id
   =imaginal>
     isa typing-word
     word =word
==>
  !bind! =char (subseq =word 0 1)
  !bind! =step *typ_step*
  +imaginal>
    isa typing-pos
    character =char
    pos 0
    step =step
    id =id
  !bind! =x (find-location-x =char) 
  !bind! =y (find-location-y =char)
  +visual-location>
    isa visual-location
    kind button
    screen-x =x
    screen-y =y
  =goal>
    state enter-hard
    vis-sp =word
    step =step
)

(p locate-next
   !eval! (typing-enabled *exp*)
   =goal>
     isa typing
     state find-next
     vis-sp =word

  =retrieval>
    isa order
    identity =pos
    next =nxt
  ?manual>
    state free
==>
  !bind! =char (if (< =pos 10) (subseq =word =pos =nxt) (subseq =word 9 10))
  !bind! =step *typ_step*

  !bind! =x (find-location-x =char) 
  !bind! =y (find-location-y =char)

  =retrieval>
  +visual-location>
    isa visual-location
    kind button
    screen-x =x
    screen-y =y
  =goal>
    state enter-hard
    step =step
)


(p hard-click-find
   !eval! *typing-on*
   =goal>
     isa typing
     state enter-hard
   =visual-location>
     isa visual-location
     kind button
   ?visual>
     state free
     buffer empty
   ?manual>
     state free
==>
  +visual>
    isa move-attention
    screen-pos =visual-location
  +manual>
    isa move-cursor
    loc =visual-location
)

(p hard-click-click1
   !eval! *typing-on*
  =goal>
    isa typing
    state enter-hard
    vis-sp =word
    step =step
  =visual>
    isa button
    value =char
  ?manual>
    state free
  =imaginal>
    isa typing-pos
    pos 0
  ?visual-location>
    buffer empty

 ==>
   +manual>
     isa click-mouse
  =imaginal>
    pos 1
  +visual>
     isa clear
  =goal>
    state next-number-hard
)

(p hard-click-click
   !eval! *typing-on*
  =goal>
    isa typing
    state enter-hard
    vis-sp =word
    step =step
  =visual>
    isa button
    value =char
  ?manual>
    state free
  =imaginal>
    isa typing-pos
  ?visual-location>
    buffer empty
  =retrieval>
    isa order
    identity =pos
    next =nxt
 ==>
   +manual>
     isa click-mouse
  =imaginal>
    character =char
    pos =nxt
    step =step
   +visual>
     isa clear
  =goal>
    state next-number-hard
)

(p next-letter-after-number-hard
   =goal>
     isa typing
     state next-number-hard
     vis-sp =word
     step =step
  =imaginal>
     isa typing-pos
     character =char
     pos =pos
   ?visual>
     state free
==>
  !bind! =id *trial* ; =id (random 10000)
  !bind! =step *typ_step*
  +visual>
    isa clear
   +imaginal>
    isa typing-word
    word =word
    id =id
    step =step
   =goal>
     state almost-next-letter
     id =id
 ;    step =step
)

(p switch-to-next-letter
   !eval! *subtraction-on*
   =goal>
     isa typing
     state almost-next-letter
   ?imaginal>
     state free
==>
   =goal>
     state next-letter
)



(p start-after-subtraction-with-right-imaginal
   !eval! *typing-on*
   !eval! (typing-stimulus *exp*)
   =goal>
     isa typing
     state next-letter
     id =id
     step =step
  =imaginal>
    isa typing-word

   ?retrieval>
    state free
    buffer empty
  ; ?manual>
  ;   state free
 ==>
   =imaginal>
  =goal>
    state find-next2
  +retrieval>
    isa typing-pos
    step =step
)

(p restore-imaginal-typing-hard
  =goal>
    isa typing
    state retrieving
    step =step
    id =id
  =retrieval>
    isa typing-word
  ?imaginal>
    state free
==>
  +imaginal>
    =retrieval
  =goal>
    state find-next2
  +retrieval>
    isa typing-pos
    step =step
)



(p check-spelling2
   !eval! (typing-enabled *exp*)
   =goal>
     isa typing
     state find-next2
     step =step
   =imaginal>
     isa typing-word
     word =wrd
   =retrieval>
     isa typing-pos
     pos =pos
 ;  ?manual>
 ;    state free
==>
  =goal>
    vis-sp =wrd
    state find-next
  +imaginal>
    =retrieval
  +retrieval>
    isa order
    identity =pos
)


(p start-after-subtraction-with-wrong-imaginal1
   !eval! *typing-on*
   !eval! (typing-stimulus *exp*)
   =goal>
     isa typing
     state next-letter
     id =id
     step =step
   =imaginal>
     isa typing-pos
   ?visual>
     state free
     buffer empty
   ?retrieval>
    state free
    buffer empty
   ?imaginal>
     state free
  ; ?manual>
  ;   state free
 ==>
   =goal>
     state retrieving
   +retrieval>
     isa typing-word
     id =id
     step =step
)

(p start-after-subtraction-with-wrong-imaginal2
   !eval! *typing-on*
   !eval! (typing-stimulus *exp*)
   =goal>
     isa typing
     state next-letter
     id =id
     step =step
   =imaginal>
     isa carry-subtraction
   ?visual>
     state free
     buffer empty
   ?retrieval>
    state free
    buffer empty
   ?imaginal>
     state free
  ; ?manual>
  ;   state free
 ==>
   =goal>
     state retrieving
   +retrieval>
     isa typing-word
     id =id
     step =step
)



;;; Subtraction
;;;;;;;;;;;;;;;;;;;

;;DM

(chunk-type sub-fact termA termB diff)
(chunk-type add-fact addA addB add)


(add-dm
 (reading isa chunk)
 (respond isa chunk)
 (find-next isa chunk)
 (find-new-letter isa chunk)
 (busy-typing isa chunk)
 (retrieving isa chunk)
 (wait-for-carry isa chunk)
 (hard isa chunk)
 (read-b-again isa chunk)
)

;;11 - 0tot9
(dotimes (j 11)
  (dotimes (i (+ 10))
    (let ((name (intern (string-upcase (concatenate 'string "sub" (write-to-string j) "-" (write-to-string i)))))
          (a (write-to-string j))
          (b (write-to-string i))
          (diff (write-to-string (- j i))))
      (when (> (- j i) -1)
          (eval `(add-dm (,name isa sub-fact termA ,a termB ,b diff ,diff)))
          (eval `(sdp ,name :creation-time -1e12 :references 15000000)))
      (when (< (- j i) 0)
          (eval `(add-dm (,name isa sub-fact termA ,a termB ,b diff ,diff)))
          (eval `(sdp ,name :creation-time -1e12 :references 130000))))))



;;19 - 0tot9
(dotimes (j 9)
  (dotimes (i (+ 10))
    (let* ((j (+ j 11))
           (name (intern (string-upcase (concatenate 'string "sub" (write-to-string j) "-" (write-to-string i)))))
           (a (write-to-string j))
           (b (write-to-string i))
           (diff (write-to-string (- j i))))
      (eval `(add-dm (,name isa sub-fact termA ,a termB ,b diff ,diff)))
      (eval `(sdp ,name :creation-time -1e12 :references 130000))))) ;Jelmer: 80000, Niels: 250000


(dotimes (j 10)
  (let ((name (intern (string-upcase (concatenate 'string "add" (write-to-string j) "+" (write-to-string 10)))))
        (a (write-to-string j))
        (b (write-to-string 10))
        (addi (write-to-string (+ j 10))))
    (eval `(add-dm (,name isa add-fact addA ,a addB ,b add ,addi)))
    (eval `(sdp ,name :creation-time -1e12 :references 1000000)))) ;last 200000

;;Production Rules

(p subtraction-start
   !eval! (subtraction-stimulus *exp*)
   !eval! *subtraction-on*
   =goal>
     isa subtraction
     state start
   ?visual-location>
     buffer empty
   ?visual>
     buffer empty
     state free
   ?imaginal>
     state free
   ?retrieval>
     state free
   ;?manual>
   ;  state free
==>
  =goal>
    state reading
  +visual-location>
    isa visual-location
    kind text
    screen-x 205
    screen-y 135
)

(p read-term-A
   !eval! *subtraction-on*
   =goal>
     isa subtraction
     state reading
   =visual-location>
     isa visual-location
     kind text
     screen-y 135
     screen-x =x
   ?visual>
      state free
==>
  +visual>
    isa move-attention
    screen-pos =visual-location
  +visual-location>
    isa visual-location
    kind text
    screen-x =x
    screen-y 195
)

(p read-term-B
   !eval! *subtraction-on*
   =goal>
     isa subtraction
     state reading
     type easy
   ?visual>
      state free
   =visual>
     isa text
     value =txt
   =visual-location>
     isa visual-location
     screen-y 195
   
==>
   +visual>
     isa move-attention
     screen-pos =visual-location
   =goal>
     state retrieving
     last =txt
)

(p start-subtraction-retrieval
   !eval! *subtraction-on*
  =goal>
    isa subtraction
    state retrieving
    type easy
    last =termA
  =visual>
    isa text
    value =termB
  ?retrieval>
    state free
    buffer empty
==>
  =visual>
  +retrieval>
    isa sub-fact
    termA =termA
    termB =termB
  =goal>
    state respond
)

(p subtraction-response-initiate
   !eval! *subtraction-on*
   =goal>
     isa subtraction
     state respond
   =retrieval>
     isa sub-fact
     termA =a
     termB =b
     diff =txt
   ?manual>
     state free
   =visual>
     isa text
  !eval! (and (> (parse-integer =txt) -1) (< (parse-integer =txt) 10)) 
==>
  =visual>
  !bind! =x (find-location-x =txt)
  !bind! =y (find-location-y =txt)
  +visual-location>
    isa visual-location
    kind button
    screen-x =x
    screen-y =y
)

(p subtraction-response-initiate-10
   !eval! *subtraction-on*
   =goal>
     isa subtraction
     state respond
   =retrieval>
     isa sub-fact
     termA =a
     termB =b
     diff =txt
   ?manual>
     state free
   =visual>
     isa text
  !eval! (equalp =txt "10") 
==>
  =visual>
  !bind! =x (find-location-x "9")
  !bind! =y (find-location-y "9")
  +visual-location>
    isa visual-location
    kind button
    screen-x =x
    screen-y =y
)

(p subtraction-response-click-find
   !eval! *subtraction-on*
   =goal>
     isa subtraction
     state respond
   =visual-location>
     isa visual-location
     kind button
   =visual>
     isa text
   ?manual>
     state free
   ?visual>
     state free
==>
  +visual>
    isa move-attention
    screen-pos =visual-location
  +manual>
    isa move-cursor
    loc =visual-location
)

(p subtraction-response-click-easy
   !eval! *subtraction-on*
  =goal>
    isa subtraction
    state respond
    type easy
  =visual>
    isa button
  ?manual>
    state free
  ?visual>
    state free
 ==>
   =goal>
     state busy-typing
   +manual>
     isa click-mouse
   +visual>
     isa clear
)

(p subtraction-response-click-hard
   !eval! *subtraction-on*
  =goal>
    isa subtraction
    state respond
    type hard
  =visual>
    isa button
  ?manual>
    state free
  ?visual>
    state free
 ==>
   =goal>
     state find-next
   +manual>
     isa click-mouse
   +visual>
     isa clear
)



(p subtraction-see-response-hard
   =goal>
     isa subtraction
     state find-next
     type hard
   =imaginal>
     isa carry-subtraction
     carry =cry
     termB =tB
     step =step
  ?visual>
    state free
==>
  !bind! =id *trial* ;(random 10000)
  !bind! =step *sub_step*
   +imaginal>
     isa carry-subtraction
     carry =cry
     termB =tB
     step =step
     id =id
   =goal>
     state busy-typing
     id =id
     step =step

)

(p subtraction-see-response-but-disabled
   !eval! (not *subtraction-on*)
  =goal>
     isa subtraction
     state busy-typing
   ?imaginal>
     state free
==>
  =goal>
    state find-new-letter
)

(p subtraction-see-response-but-disabled-signal-no-ima
   !eval! (not *subtraction-on*)
   !eval! *signal*
  =goal>
     isa subtraction
     state find-next
   ?imaginal>
     state free
     buffer empty
==>
  =goal>
    state find-new-letter
)

(p subtraction-see-response-but-disabled-signal-wrong-ima
   !eval! (not *subtraction-on*)
   !eval! *signal*
  =goal>
     isa subtraction
     state find-next
   ?imaginal>
     state free
   =imaginal>
     isa typing-word
==>
  =imaginal>
  =goal>
    state find-new-letter
)


(p subtraction-find-current-letter-easy
   !eval! *subtraction-on*

 =goal>
     isa subtraction
     state find-new-letter
     type easy
   ?visual-location>
   - state error
   ?visual>
     state free
   ?imaginal>
     state free
   ?retrieval>
     state free
  ; ?manual>
   ;  state free
==>
  +visual-location>
    isa visual-location
    kind text
    screen-x 205
    screen-y 135
  =goal>
    state reading
)

(p subtraction-find-current-letter-hard-imaginal
   !eval! *subtraction-on*
   !eval! (busy *exp*)
  ; !eval! (not *signal*) ;switch this on for purely visual carry strategy, and comment out the ima use in the rule below
   =goal>
     isa subtraction
     state find-new-letter
     type hard
   ?visual-location>
   - buffer full
   - state error
   ?visual>
     state free
     buffer empty
   =imaginal>
     isa carry-subtraction
   ?imaginal>
     state free
   ?retrieval>
     state free
 ;  ?manual>
 ;    state free
==>
   !bind! =step *sub_step*
   =goal>
     step =step
  =imaginal>
    step =step
  +visual-location>
    isa visual-location
    kind text
    screen-x 205
    screen-y 135
  =goal>
    state reading
)

(p subtraction-find-current-letter-hard-signal-no-ima
   !eval! *subtraction-on*
   !eval! (busy *exp*)
   !eval! *signal*
   =goal>
     isa subtraction
     state find-new-letter
     type hard
   ?visual-location>
   - buffer full
   - state error
   ?visual>
     state free
     buffer empty
   ?imaginal>
     state free
     buffer empty
 ;  ?manual>
  ;   state free
==>
  +visual-location>
    isa visual-location
    kind text
    screen-x 235
    screen-y 100
  =goal>
    state encode-carry
)

(p subtraction-find-current-letter-hard-signal-wrong-ima1
   !eval! *subtraction-on*
   !eval! (busy *exp*)
   !eval! *signal*
   =goal>
     isa subtraction
     state find-new-letter
     type hard
   ?visual-location>
   - buffer full
   - state error
   ?visual>
     state free
     buffer empty
   =imaginal>
     isa typing-word
  ; ?manual>
  ;   state free
==>
  =imaginal>
  +visual-location>
    isa visual-location
    kind text
    screen-x 235
    screen-y 100
  =goal>
    state encode-carry
)

(p subtraction-find-current-letter-hard-signal-wrong-ima2
   !eval! *subtraction-on*
   !eval! (busy *exp*)
   !eval! *signal*
   =goal>
     isa subtraction
     state find-new-letter
     type hard
   ?visual-location>
   - buffer full
   - state error
   ?visual>
     state free
     buffer empty
   =imaginal>
     isa typing-pos
  ; ?manual>
  ;   state free
==>
  =imaginal>
  +visual-location>
    isa visual-location
    kind text
    screen-x 235
    screen-y 100
  =goal>
    state encode-carry
)

(p encode-carry-signal
   !eval! *subtraction-on*
   !eval! *signal*
   =goal>
     isa subtraction
     state encode-carry
   =visual-location>
     isa visual-location
     screen-x 235
   ?visual>
     state free
==>
   +visual>
     isa move-attention
     screen-pos =visual-location
)

(p parse-carry-signal
   !eval! *subtraction-on*
   !eval! *signal*
   =goal>
     isa subtraction
     state encode-carry
   =visual>
     isa text
     value =val
   ?imaginal>
     state free
==>
   !bind! =cry (if (equalp =val "1") 1 0)
   !bind! =id *trial* ; (random 10000)
   !bind! =step *sub_step*
  +imaginal>
     isa carry-subtraction
     carry =cry
     id =id
     step =step
     termb 0
   =goal>
     state reading
  +visual-location>
    isa visual-location
    kind text
    screen-x 205
    screen-y 135
)



(p subtraction-find-current-letter-hard-wrong-imaginal1
   !eval! *subtraction-on*
   !eval! (not *signal*)
  =goal>
     isa subtraction
     state find-new-letter
     type hard
     id =id
     step =step
   ?visual-location>
   - buffer full
   - state error
   ?visual>
     state free
     buffer empty
   =imaginal>
    isa typing-word
   ?retrieval>
     state free
     buffer empty
   ?imaginal>
     state free
   ;?manual>
  ;   state free
==>
   +retrieval>
     isa carry-subtraction
     id =id
     step =step
   ;+visual>
   ;  isa move-attention
   ;  screen-pos task-left
)

(p subtraction-find-current-letter-hard-wrong-imaginal2
   !eval! *subtraction-on*
   !eval! (not *signal*)
  =goal>
     isa subtraction
     state find-new-letter
     type hard
     id =id
     step =step
   ?visual-location>
   - buffer full
   - state error
   ?visual>
     state free
     buffer empty
   =imaginal>
    isa typing-pos
   ?retrieval>
     state free
     buffer empty
   ?imaginal>
     state free
   ;?manual>
  ;   state free
==>
   +retrieval>
     isa carry-subtraction
     id =id
     step =step
   ;+visual>
   ;  isa move-attention
   ;  screen-pos task-left
)

(p subtraction-restore-imaginal
   =goal>
     isa subtraction
     type hard
   =retrieval>
    isa carry-subtraction
    carry =cry
    termb =tb
 ==>
   !bind! =id *trial* ; (random 10000)
   !bind! =step *sub_step*
   =goal>
     id =id
     step =step
   +imaginal>
     isa carry-subtraction
     carry =cry
     step =step
     id =id
     termb =tb
)

(p subtraction-wow-it-is-hard
   !eval! *subtraction-on*
   =goal>
     isa subtraction
     state respond
     type easy
   =retrieval>
     isa sub-fact
     termA =a
     termB =b
     diff =diff
  !eval! (< (parse-integer =diff) 0)
==>
   !bind! =id *trial* ; =id (random 10000)
   !bind! =step *sub_step*
   =goal>
     state add-tens
     type hard
     id =id
     step =step
     last =a
   +imaginal>
     isa carry-subtraction
     carry 0
     termB =b
     step =step
     id =id
  +visual-location>
     isa visual-location
     kind text
     screen-y 135
     screen-x current
  !eval! (format t "help it's hard!")
)

(p subtraction-hard-addition
   !eval! *subtraction-on*
   =goal>
     isa subtraction
     state respond
     type hard
   =retrieval>
     isa sub-fact
     termA =a
     termB =b
     diff =diff
   !eval! (< (parse-integer =diff) 0)
   =imaginal>
     isa carry-subtraction
     carry =cr
   =visual>
     isa text
     screen-pos =pos
==>
 !bind! =id *trial* ; =id (random 10000)
   !bind! =step *sub_step*
   =goal>
    last =a
    id =id
    step =step
    state add-tens  
  =imaginal>  ;;carry not updated yet, just used to remember =b
     carry =cr
     termB =b
     id =id
     step =step
  +visual-location>
     isa visual-location
     kind text
     screen-y 135
     screen-x current
)

(p subtraction-hard-addition-signal
   !eval! *subtraction-on*
   !eval! *signal*
   =goal>
     isa subtraction
     state respond
     type hard
   =retrieval>
     isa sub-fact
     termA =a
     termB =b
     diff =diff
   !eval! (< (parse-integer =diff) 0)
   =visual>
     isa text
     screen-pos =pos
==>
 !bind! =id *trial* ; =id (random 10000)
   !bind! =step *sub_step*
   =goal>
    last =a
    id =id
    step =step
    state add-tens  
  +visual-location>
     isa visual-location
     kind text
     screen-y 135
     screen-x current
)


(p add-tens-look-up
   !eval! *subtraction-on*
   =goal>
     isa subtraction
     state add-tens
     type hard
     last =a
   ?imaginal>
     state free
     buffer full
   =visual-location>
     isa visual-location
     kind text
     screen-y 135
     screen-x =x
   ?visual>
     state free
==>
  +visual>
    isa move-attention
    screen-pos =visual-location
  =imaginal>
)

(p add-tens-look-up-signal
   !eval! *subtraction-on*
   !eval! *signal*
   =goal>
     isa subtraction
     state add-tens
     type hard
     last =a
   =visual-location>
     isa visual-location
     kind text
     screen-y 135
     screen-x =x
   ?visual>
     state free
==>
  +visual>
    isa move-attention
    screen-pos =visual-location
)



(p addition-done-time-for-subtraction
   !eval! *subtraction-on*
   =goal>
     isa subtraction
     state add-tens
     type hard
     last =a
   =visual>
    isa text
   =imaginal>
     isa carry-subtraction
     termB =b
     id =id
     step =step
   ?visual>
     state free
==>

  !bind! =string-add (write-to-string (+ (values (parse-integer =a)) 10))
  +imaginal> ;now we update the imaginal
    isa carry-subtraction
    termB =b
    carry 1
    id =id
    step =step
  =goal>
    state read-b-again
    last =string-add
  +visual-location>
    isa visual-location
    screen-x current
    screen-y 195
)


(p addition-done-time-for-subtraction-signal
   !eval! *subtraction-on*
   !eval! *signal*
   =goal>
     isa subtraction
     state add-tens
     type hard
     last =a
   =visual>
    isa text
   ?visual>
     state free
==>

   !bind! =string-add (write-to-string (+ (values (parse-integer =a)) 10))
   !bind! =id *trial* ; =id (random 10000)
   !bind! =step *sub_step*

  +imaginal> ;now we update the imaginal
    isa carry-subtraction
    termB 0
    carry 1
    id =id
    step =step
  =goal>
    state read-b-again
    last =string-add
  +visual-location>
    isa visual-location
    screen-x current
    screen-y 195
)


(p read-b-again-subtraction-hard
   !eval! *subtraction-on*
   =goal>
     isa subtraction
     state read-b-again
     type hard
   =visual-location>
     isa visual-location
     kind text
     screen-y 195
   ?visual>
     state free
==>
  +visual>
    isa move-attention
    screen-pos =visual-location
  =goal>
    state retrieving
)

(p start-subtraction-retrieval-hard
   !eval! *subtraction-on*
  =goal>
    isa subtraction
    state retrieving
    type hard
    last =termA
  =visual>
    isa text
    value =termB
 ?retrieval>
    state free
    buffer empty
  =imaginal>
    isa carry-subtraction
==>

  =visual>
  =imaginal>

  +retrieval>
    isa sub-fact
    termA =termA
    termB =termB
  =goal>
    state respond
)

(p start-subtraction-retrieval-hard-signal
   !eval! *subtraction-on*
   !eval! *signal*
  =goal>
    isa subtraction
    state retrieving
    type hard
    last =termA
  =visual>
    isa text
    value =termB
 ?retrieval>
    state free
    buffer empty
==>

  =visual>
  +retrieval>
    isa sub-fact
    termA =termA
    termB =termB
  =goal>
    state respond
)


(p read-term-B-hard-no-imaginal-first-step-no-signal
   !eval! *subtraction-on*
   !eval! (not *signal*)
   =goal>
     isa subtraction
     state reading
     type hard
   =visual>
     isa text
     value =txt
   =visual-location>
     isa visual-location
     screen-y 195
   ?imaginal>
     state free
     buffer empty
   ?visual>
     state free
   ?retrieval>
     state free
     buffer empty
==>
   !bind! =id *trial* ; =id (random 10000)
   !bind! =step *sub_step*
   +imaginal>
     isa carry-subtraction
     carry 0
     id =id
     termB 0
     step =step
   +visual>
     isa move-attention
     screen-pos =visual-location
   =goal>
     state retrieving
     last =txt
     id =id
     step =step
)

(p read-term-B-hard-no-imaginal-first-step-signal
   !eval! *subtraction-on*
   !eval! *signal*
   =goal>
     isa subtraction
     state reading
     type hard
   =visual>
     isa text
     value =txt
   =visual-location>
     isa visual-location
     screen-y 195
   ?imaginal>
     state free
     buffer empty
   ?visual>
     state free
   ?retrieval>
     state free
     buffer empty
==>
   !bind! =id *trial* ; =id (random 10000)
   !bind! =step *sub_step*
   +visual>
     isa move-attention
     screen-pos =visual-location
   =goal>
     state retrieving
     last =txt
     id =id
     step =step
)


(p read-term-B-hard-no-carry
   !eval! *subtraction-on*
   =goal>
     isa subtraction
     state reading
     type hard
   =visual>
     isa text
     value =txt
   =visual-location>
     isa visual-location
     screen-y 195
   =imaginal>
     isa carry-subtraction
     carry 0
   ?visual>
     state free
==>
   =imaginal>
   +visual>
     isa move-attention
     screen-pos =visual-location
   =goal>
     state retrieving
     last =txt
)

(p read-term-B-hard-with-carry
   !eval! *subtraction-on*
   =goal>
     isa subtraction
     state reading
     type hard
   =visual>
     isa text
     value =txt
   =visual-location>
     isa visual-location
     screen-y 195
   =imaginal>
     isa carry-subtraction
     carry 1
   ?visual>
     state free
   ?retrieval>
     state free
     buffer empty
==>
   =imaginal>
   +visual>
     isa move-attention
     screen-pos =visual-location
   +retrieval>
     isa sub-fact
     termA =txt
     termB "1"
   =goal>
     state wait-for-carry  
)

(p process-carry
   !eval! *subtraction-on*
   =goal>
     isa subtraction
     state wait-for-carry
     type hard
   =retrieval>
     isa sub-fact
     termA =a
     diff =diff
   =imaginal>
     isa carry-subtraction
==>
  =imaginal>
    carry 0
  =goal>
    state retrieving
    last =diff
)


)

